
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <signal.h>
#include <math.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h>

/* macros */
#define _VERBOSE1 if(verbose >= 1)
#define _VERBOSE2 if(verbose >= 2)
#define _VERBOSE3 if(verbose >= 3)
#define _VERBOSE4 if(verbose >= 4)
#define _VERBOSE5 if(verbose >= 5)
#define _VERBOSE6 if(verbose >= 6)
#define _VERBOSE7 if(verbose >= 7)

/**** structures ****/

/* describing a job and it's state */
struct job {
 /* selected image range */
 gint start_frame;
 gint end_frame;
 /* start/end-frame for this job */
 gint job_start_frame;
 gint job_end_frame;
 gint job_current_frame;
 /* true if running */
 gboolean active;
 /* true if finished */
 gboolean finished;
 /* progressbar */
 GtkWidget *progressbar;
 /* progress of job (between 0-1, 1 means done) */
 gdouble progress;
 /* pointer to job that has to
	run before this job (or NULL) */
 struct job *needs_job;
 /* node this job runs on */
 struct grid_node *job_node;
 /* imageset to work on */
 struct imageset *image_set;
 /* effect */
 struct effect *job_effect;
 /* effect-data, this is a snapshot-copy */
 void *effect_data;
 /* process pid */
 pid_t process_pid;
 /* fifo pipes */
 int process_fifo_in[2], process_fifo_out[2];
};
/* this is a glue structure for effects */
struct effect {
 /* effect id */
 int effect_id;
 /* effect name */
 char *effect_name;
 /* called when current-frame changes
	and effect tab is visible */
 void (*frame_update_function)();
 /* validate the gui (enable/disable buttons etc) */
 void (*update_gui)(struct effect *this_effect);
 /* function to generate command string */
 char *(*process_function)(struct job *this_job);
 /* called before another 
	effect-tab gets visible */
 void (*exit_function)();
 /* returns true if effect is selected */
 int (*selected_function)();
 /* image dimensions (when image loaded) */
 gint image_width;
 gint image_height;
 /* currently processed frame */
 int current_frame;
 /* size of effect-specific structure */
 int effect_data_size;
 /* effect specific data */
 void *effect_data;
};
/* this describes one imageset */
struct imageset {
 /* unique name of imageset and directory prefix*/
 char *name;
 /* amount of frames in this set */
 int total_frames;
 /* current effect */
 struct effect *current_effect;
 /* list with pointers to effects */
 GSList *effects;
 /* current frame shown */
 int current_frame;
 /* selected image range */
 int start_frame;
 int end_frame;
 /* orginal dimensions */
 int frame_width;
 int frame_height;
 /* current image */
 GdkPixbuf *current_image;
 /* previous image */
 GdkPixbuf *previous_image;
 /* next image */
 GdkPixbuf *next_image;
 /* preview drawingareas */
 GtkWidget *drawingarea_pre;
 GtkWidget *drawingarea_current;
 GtkWidget *drawingarea_next;
 /* label of imageset */
 GtkWidget *label;
 /* current list of images */
 GSList *image_list;
 /* list of jobs for this imageset */
 GSList *jobs;
 /* scale factor of pre- 
 current- and next-preview */
 double scale_factor;
 /* offsets of pre- 
 current- and next-preview */
 int x_offset, y_offset;
 /* offsets of crop-preview */
 int x_crop_offset, y_crop_offset;
 /* clear workingdir if TRUE */
 char *cleardir;
};

/* a processing node - will be part of a GSlist */
struct grid_node {
 /* ip or hostname */
 gchar *address;
 /* login username */
 gchar *username;
 /* working directory */
 gchar *directory;
 /* amount of processes to start on this node */
 gint processes;
 /* amount of processes running */
 gint processes_running;
 /* FALSE if node should be ignored */
 gboolean enabled;
 /* list of jobs for this node */
 GSList *job_queue;
};

struct cmd_morph_constant {
 gint iterations;
};
struct cmd_morph_nonlinear {
 gint iterations_max;
 int vectors;
 gfloat *vector;
};
struct cmd_crop_constant {
 gint top;
 gint bottom;
 gint left;
 gint right;
 gint bg_color;
};
struct cmd_scale_constant {
 /* new dimensions */
 gint new_width;
 gint new_height;
 /* new dimensions percentage */
 gdouble percent_width;
 gdouble percent_height;
 /* scale factor */
 gdouble w_factor;
 gdouble h_factor;
 /* TRUE if we prevent aspect ratio */
 gint aspect;
};

/* structure to hold dimensions
 of an image, scaled to a specific drawingarea */
struct fitted_image {
 /* drawingarea to fit image in */
 GtkWidget *drawingarea;
 /* source dimensions */
 int image_width;
 int image_height;
 /* dest dimensions */
 int scaled_width;
 int scaled_height;
 /* scale factors */
 double w_factor;
 double h_factor;
 /* offset inside drawingarea */
 int x_offset;
 int y_offset;
} current_fitting;


/* list of columns for preferences list */
enum {
 COL_ADDRESS = 0,
 COL_USERNAME,
 COL_PROCESSES,
 COL_ACTIVE,
 NUM_COLS
};

/* modelist for grid_generate_ssh() */
enum {
 SSH_CURRENT = 0,
 SSH_NEXT,
};

/************** strings **************/
char *config_filename;

/**** functions ****/
int parse_arguments(int argc, char* argv[]);
void import_images(GSList *images);
void free_imageset(struct imageset *images);
gint draw_preview(gchar *path, gint field);
void update_gui();
void update_preview_images(int image);
void update_effect_crop_image();
void update_effect_crop_lines();
void update_effect_scale();
gint render_fade_constant(struct cmd_morph_constant *command);
gint render_fade_nonlinear(struct cmd_morph_nonlinear *command);
gint render_crop(struct effect *thisEffect);
gint render_scale(struct cmd_scale_constant *command);	
void fit_image(struct fitted_image *fitting);
void alert_dialog(gchar *message);
void progressbar_set(GtkWidget *progressbar, gdouble fraction);
void nonlinear_ruler_update();
gint get_processed_frames_amount();
void print_help(char *processname);
gint draw_image(gchar *path, GtkWidget *drawingarea, GdkPixmap *pixmap);
int mplayer_init(char *mplayer_path, char *videoname, char *exportcmd, char *pathname);
int mencoder_init(char *listfile, char *video_dir);
void mencoder_cleanup();
gint export_video(GSList *imagelist, gchar *video_dir, gchar *video_file);
gint export_video_mencoder(int amount);
gint import_video_mplayer(gchar *imagedir);
int mplayer_cleanup();
void mplayer_sendcommand(char *cmd);
void put_status(char *text);
gint sort_images(char *a, char *b);
gint sort_sub_images(char *a, char *b);
gboolean dir_isempty(char *dirname);
void alert_cleardir(char *dirname);
void cb_cleardir_show(GtkWidget *button);
void cb_preview_scale_post_changed(GtkWidget *drawingarea);
void cb_preview_pre_changed(GtkWidget *drawingarea);
void cb_preview_current_changed(GtkWidget *drawingarea);
void cb_preview_next_changed(GtkWidget *drawingarea);
gint cb_pre_exposed(GtkWidget *drawingarea, GdkEventExpose *event);
gint cb_current_exposed(GtkWidget *drawingarea, GdkEventExpose *event);
gint cb_next_exposed(GtkWidget *drawingarea, GdkEventExpose *event);
void cb_import_video_dialog();
void cb_import_dialog(GtkWidget *menu);
void cb_scale_percentage_changed(GtkWidget *spinbutton);
void cb_imageset_name_dialog();
void cb_import_videodir_dialog(GtkWidget *menu);
void cb_effect_changed(GtkNotebook *notebook, GtkNotebookPage *page, guint pagenum);
void new_imageset_tab(struct imageset *new_set);
gint play_video_mplayer(gchar *path);
void init_versions();
void update_effect_scale_images();
void init_effect_scale(struct effect *scale_effect);
void exit_effect_scale(struct effect *scale_effect);
void draw_fitted_image(GdkPixbuf *image, GdkPixmap *dest, GtkWidget *drawingarea);
void effect_scale_set_labels();
void free_all_imagesets(GSList *imagesets);
void recall_func_store(void *function);
void recall_func_run();
void remove_imageset_tab(int tabnum);
void morph_update_gui();
void init_effects();
void init_prefs_grid();
void free_effects();
void free_grid_nodes();
void effect_free(struct effect *old_effect);
void effect_morph_constant_bootstrap();
void effect_morph_nonlinear_bootstrap();
void effect_scale_bootstrap();
void effect_crop_bootstrap();
void grid_add_node(gchar *address, gchar *user, gchar *directory, gint jobs, gint enabled);
gboolean grid_remove_nth_row (gint n);
void grid_create_node(gchar *address, gchar *user, gchar *directory, gint jobs, gint enabled);
void grid_job_cleanup(struct job *this_job);
char *grid_generate_ssh(struct job *this_job, char *effect_command, int mode);
void grid_job_launch(struct job *this_job);
void prefs_init();
void prefs_free();
char *prefs_get(char *nodename, char *propname);
int prefs_create_config();
void reset_imageset_name();
void cb_nonlinear_curve_changed(GtkWidget *curve);
void update_crop_properties();
void free_imagelist(GSList *images);

/************* globals *************/
/* xml file for our gui */
GladeXML *xml;
/* set verbosity level */
int verbose;
/* list of effect-init routines */
typedef void function(void);

/* always remember the previewing DrawingAreas */
GtkWidget *drawingarea_pre, *drawingarea_current, *drawingarea_next;

/* amount of jobs done in one marshalling session */
gdouble total_jobs_done;
/* list of effect structures */
GSList *effects;
/* list of imagesets */
GSList *imagesets;
/* list of grid computing nodes */
GSList *grid_nodes;
/* imagelist clipboard */
GSList *clipboard_image_range;

/* current imageset */
struct imageset *current_imageset;
/* TRUE if node-marshall is running */
gboolean marshall_running;

/* TRUE if user wants to cancel
	an operation */
gboolean flag_cancel;
/* TRUE if mplayer is active */
gboolean flag_import;
/* TRUE if video-export process active */
gboolean flag_export;
/* fraction of the main progress-bar */
double current_fraction;
/* function that was interrupted and some other
  user-interaction function has to finish until
  we continue - singly linked list of functions */
GSList *recall_function;
void (*recall_func)();
char *video_importdir;

/* video export fifo pipes */
int export_fifo_in[2], export_fifo_out[2];
FILE *export_fifo;
pid_t export_pid;

/* doublebuffers */
GdkPixmap *preview_pre, *preview_current, *preview_next;
GdkPixmap *preview_crop;
GdkPixmap *preview_scale;

/* xml stuff */
xmlDocPtr	prefs_doc;
xmlDocPtr	project_doc;
xmlParserCtxtPtr prefs_parser;
xmlParserCtxtPtr project_parser;
char *prefs_xmlBuffer;
char *project_xmlBuffer;

/* mplayer process pid */
pid_t mplayer_pid;
/* true if mplayer running */
gboolean flag_mplayer;
/* file descriptors for fifos */
FILE *fifo_mplayer_to;
FILE *fifo_mplayer_from;
/* fifo pipes */
int mplayer_fifo_in[2], mplayer_fifo_out[2];
 
/* different effects */
#define EFFECT_MORPH_CONSTANT 0
#define EFFECT_MORPH_NONLINEAR 1
#define EFFECT_CROP 2
#define EFFECT_SCALE 3
